% 
close all 
clear all

set(0,'defaultTextFontName', 'Arial');
set(0,'defaultTextFontSize', 14);
set(0,'defaultTextFontWeight', 'Bold');
set(0,'defaultTextBackgroundColor', [1,1,1]);
set(0,'defaultLineLineWidth', 3);

flags.noAnimation = false;
flags.drawFloor = true;
flags.drawVanishing = true;
flags.drawAxis = false;


%% robot specification
% robot robot, used as inputs for the system
robot.dt = 0.1;
robot.baseSpeed = 0.5; % [m/s], used to generate motion

% information used for robot drawings

robot.shape2D = [
    .6       -.3       -.3       .6
    0       0.25     -0.25    0
    1       1       1       1
    ];

robot.salientPoint1 = [-.3 ;0 ;1];
robot.salientPoint2 = [.3 ;0 ;1];

robot.color = [0 .7 0];

% robot complete motion
robot.vw_inputs = [ ...
    rectMovementBaseSpeed(.5, robot.baseSpeed,robot.dt); ...
    arcMovementBaseSpeed(.5, 2*pi, robot.baseSpeed , robot.dt); ...
    arcMovementBaseSpeed(.5,  - pi, robot.baseSpeed , robot.dt); ...
    rectMovementBaseSpeed(1.5, robot.baseSpeed,robot.dt); ...
    arcMovementBaseSpeed(1, - pi, robot.baseSpeed , robot.dt); ...
    rectMovementBaseSpeed(.5, robot.baseSpeed,robot.dt); ...
    arcMovementBaseSpeed(.5, - pi / 2, robot.baseSpeed , robot.dt); ...
    rectMovementBaseSpeed(2.5, robot.baseSpeed,robot.dt); ...
    arcMovementBaseSpeed(.5, - pi, robot.baseSpeed , robot.dt); ...
    rectMovementBaseSpeed(1.5, robot.baseSpeed, robot.dt); ...
    arcMovementBaseSpeed(.5, - pi / 2, robot.baseSpeed , robot.dt); ...
    rectMovementBaseSpeed(1.0, robot.baseSpeed, robot.dt); ...
];

robot.time = (0:size(robot.vw_inputs, 1)) * robot.dt;

% robot position start
robot.init_state = [0 0 0 0 0 0]; % robot starting point (x,y,z,theta,phi,rho)
robot.pose = nan(length(robot.time),6);
robot.pose(1,:) = [robot.init_state];

%% world drawings on the floor
floor.draws = [ -2 2 2 -2 -2;
                -2 -2 2 2 -2
                0 0 0 0 0
                1 1 1 1 1];
            


%% camera specification

camera.cameraWRTworld = createMatrix(-4,-.5,2.5,0,deg2rad(40),0) * createMatrix(0,0,0,0,pi/2,0) * createMatrix(0,0,0,-pi/2,0,0);
camera.worldWRTcamera = invRT3D(camera.cameraWRTworld);

camera.size = [320; 240]; 
camera.f = 320; 
camera.cx = 160; 
camera.cy = 120; 

camera.imageLimits = [    % image limits
        0 camera.size(1) camera.size(1) 0 0
        0 0 camera.size(2) camera.size(2) 0
        1   1   1   1   1];

% camera matrix
camera.K = createKMatrixPinHole(camera.f, camera.f, camera.cx, camera.cy); 
% complete camera matrix (intrinsic + extrinsic parameters)
camera.P = createPMatrix(camera.worldWRTcamera(1:3,1:3), camera.worldWRTcamera(1:3,4), camera.K);

% this homography link a point on z = 0 plane to image plane
% pimg = Hz0 * [x;y;1], with x,y point on z=0 plane   
camera.Hz0 = camera.P * [1 0 0; 0 1 0; 0 0 0; 0 0 1]; 

% this homography link a point on image to z = 0 plane
% pz0 = invHz0 * [u;v;1], with u,v point on image plane (in pixel)
camera.invHz0 = inv(camera.Hz0);

% these are the limits of field of view on z = 0 plane
camera.worldLimits = hnormalise(camera.invHz0 * camera.imageLimits);
camera.worldLimits3D =  [camera.worldLimits(1:2,:); zeros(1,5); camera.worldLimits(3,:)];

%% first camera question: where is the camera in the world?
camera.R_worldWRTcamera = inv(camera.K) * camera.P(1:3,1:3);
camera.t_worldWRTcamera = inv(camera.K) * camera.P(1:3,4);

M = [camera.R_worldWRTcamera camera.t_worldWRTcamera; 0 0 0 1];
cameraWRTworld = invRT3D(M);
[x, y, z, rz, ry, rx] = extractPar(cameraWRTworld);
fprintf('camera is at x %f, y %f, z %f, \n\t rotation rx %f, ry %f, rz %f\n\n',x, y, z, rad2deg(rx), rad2deg(ry), rad2deg(rz));

cameraWRTworld_sensor = cameraWRTworld * createMatrix(0,0,0,pi/2,0,0) * createMatrix(0,0,0,0,-pi/2,0);
[x, y, z, rz, ry, rx] = extractPar(cameraWRTworld_sensor);
fprintf('camera (axis like world) is at x %f, y %f, z %f, \n\t rotation rx %f, ry %f, rz %f\n\n',x, y, z, rad2deg(rx), rad2deg(ry), rad2deg(rz));

%% principal axis
M = camera.P(1:3,1:3);
m = camera.P(1:3,4);
O_w = -inv(M)*m;
% d = camera.cameraWRTworld(1:3,1:3) * [0,0,1]';
d = inv(M) * [camera.cx,camera.cy,1]';
d = d/norm(d);
lambda = -O_w(3)/d(3);
O_w_axend = O_w + lambda*d;



%% robot motion

for tid = 2 : length(robot.time)
    % evolute robot system
    [robot.pose(tid,:), point1,point2] = robotMotion(robot.pose(tid - 1,:), robot.vw_inputs(tid - 1, :), robot.dt, robot.salientPoint1, robot.salientPoint2);    
    pIMG1(:,tid) = camera.Hz0 * point1([1,2,4]);
    pIMG2(:,tid) = camera.Hz0 * point2([1,2,4]);
    
    % calculate robot position by image
    pointR1 = camera.invHz0 * pIMG1(:,tid);
    pointR2 = camera.invHz0 * pIMG2(:,tid);
    d = pointR2 - pointR1;
    theta = atan2(d(2),d(1));
%     fprintf('real pose %f %f %f\n',robot.pose(tid,1),robot.pose(tid,2),robot.pose(tid,4));
%     fprintf('estimated pose %f %f %f\n',(pointR1(1)+pointR2(1))/2,(pointR1(2)+pointR2(2))/2,theta);
    assert(abs( (pointR1(1)+pointR2(1))/2 - robot.pose(tid,1)) < 1e-14);
    assert(abs( (pointR1(2)+pointR2(2))/2 - robot.pose(tid,2)) < 1e-14);
    assert(abs( (theta - robot.pose(tid,4))) < 1e-14);
end

figure()
    set(gcf, 'Position', get(0,'Screensize')); 
subplot(2,1,1)
hold on
grid on
drawWorld(camera.cameraWRTworld, camera.worldLimits3D);
plot3(robot.pose(:,1),robot.pose(:,2),robot.pose(:,3),'r');
if(flags.drawFloor)
    plot3(floor.draws(1,:), floor.draws(2,:), floor.draws(3,:),'m'); 
end

if(flags.drawAxis)
    plot3([O_w(1) O_w_axend(1)],[O_w(2) O_w_axend(2)],[O_w(3) O_w_axend(3)],'b--');
end

axis equal
view(-45,30)
ax = axis();
ax(5) = 0;
axis(ax);
hold off


% calculate complete robot pose in image
imageTraj = hnormalise( camera.P * [robot.pose(:,1:3)'; ones(1,length(robot.pose(:,1)))]);
imageFloor = hnormalise( camera.P * floor.draws);

subplot(2,1,2)

hold on

worldInImage = hnormalise(camera.P * [0 0 0 1]');
xVan = hnormalise(camera.P * [1 0 0 0]'); 
yVan = camera.P * [0 1 0 0]'; % it remains at infinity
zVan = camera.P * [0 0 1 0]'; % unuseful in this context

if(flags.drawFloor)
    plot(imageFloor(1,:), imageFloor(2,:),'m'); 
end
if(flags.drawVanishing)
    plot(xVan(1,:), xVan(2,:),'r+'); 
    plot([xVan(1,:) imageFloor(1,1)], [xVan(2,:) imageFloor(2,1)],'m--');
    plot([xVan(1,:) imageFloor(1,3)], [xVan(2,:) imageFloor(2,3)],'m--');
end

plot(worldInImage(1),worldInImage(2),'b+');

plot(imageTraj(1,1:tid),imageTraj(2,1:tid),'r.')

axis equal
axis ij
grid on
if(~flags.drawVanishing)
    axis([0 camera.size(1) 0 camera.size(2)]);  
end
hold off

%% animation
if(~flags.noAnimation)
    


figure() %3d animation
    set(gcf, 'Position', get(0,'Screensize')); 
for tid = 1 : length(robot.time)
    fprintf('iteration %d/%d\n',tid,length(robot.time));

    h1 = subplot(2,1,1);
    cla(h1);
    hold on
    drawWorld(camera.cameraWRTworld, camera.worldLimits3D);
    
    robotShape3D = drawRobot(robot.pose(tid,:),robot.shape2D); % get the shape of the robot in 3D
    
    plot3(robot.pose(tid,1),robot.pose(tid,2),robot.pose(tid,3),'r+');
    plot3(robot.pose(1:tid,1),robot.pose(1:tid,2),robot.pose(1:tid,3),'r.');
    
    axis equal
    grid on
    view(-45,30)
    axis(ax);
    hold off

    h2 =subplot(2,1,2);
    cla(h2);
    hold on

    robotShapeImage = hnormalise( camera.P * robotShape3D);
    worldInImage = hnormalise(camera.P * [0 0 0 1]');
    xDir = camera.P * [1 0 0 0]'; xDir = xDir / norm(xDir);
    yDir = camera.P * [0 1 0 0]'; yDir = yDir / norm(yDir);
    zDir = camera.P * [0 0 1 0]'; zDir = zDir / norm(zDir);
    
    fill(robotShapeImage(1,:),robotShapeImage(2,:),[.5 0 .5]);
    plot(worldInImage(1),worldInImage(2),'b+');
    
    plot(imageTraj(1,1:tid),imageTraj(2,1:tid),'r.')
    
    axis equal
    axis ij
    grid on
    axis([0 camera.size(1) 0 camera.size(2)]);
    hold off
    drawnow();

end

end